import functools

from wandb_promise import Promise, is_thenable, promise_for_dict

from ...pyutils.cached_property import cached_property
from ...pyutils.default_ordered_dict import DefaultOrderedDict
from ...type import (GraphQLInterfaceType, GraphQLList, GraphQLNonNull,
                     GraphQLObjectType, GraphQLUnionType)
from ..base import ResolveInfo, Undefined, collect_fields, get_field_def
from ..values import get_argument_values
from ...error import GraphQLError
try:
    from itertools import izip as zip
except:
    pass


def get_base_type(type):
    if isinstance(type, (GraphQLList, GraphQLNonNull)):
        return get_base_type(type.of_type)
    return type


def get_subfield_asts(context, return_type, field_asts):
    subfield_asts = DefaultOrderedDict(list)
    visited_fragment_names = set()
    for field_ast in field_asts:
        selection_set = field_ast.selection_set
        if selection_set:
            subfield_asts = collect_fields(
                context, return_type, selection_set,
                subfield_asts, visited_fragment_names
            )
    return subfield_asts


def get_resolvers(context, type, field_asts):
    from .resolver import field_resolver
    subfield_asts = get_subfield_asts(context, type, field_asts)

    for response_name, field_asts in subfield_asts.items():
        field_ast = field_asts[0]
        field_name = field_ast.name.value
        field_def = get_field_def(context and context.schema, type, field_name)
        if not field_def:
            continue
        field_base_type = get_base_type(field_def.type)
        field_fragment = None
        info = ResolveInfo(
            field_name,
            field_asts,
            field_base_type,
            parent_type=type,
            schema=context and context.schema,
            fragments=context and context.fragments,
            root_value=context and context.root_value,
            operation=context and context.operation,
            variable_values=context and context.variable_values,
        )
        if isinstance(field_base_type, GraphQLObjectType):
            field_fragment = Fragment(
                type=field_base_type,
                field_asts=field_asts,
                info=info,
                context=context
            )
        elif isinstance(field_base_type, (GraphQLInterfaceType, GraphQLUnionType)):
            field_fragment = AbstractFragment(
                abstract_type=field_base_type,
                field_asts=field_asts,
                info=info,
                context=context
            )
        resolver = field_resolver(field_def, exe_context=context, info=info, fragment=field_fragment)
        args = get_argument_values(
            field_def.args,
            field_ast.arguments,
            context and context.variable_values
        )
        yield (response_name, Field(resolver, args, context and context.context_value, info))


class Field(object):
    __slots__ = ('fn', 'args', 'context', 'info')

    def __init__(self, fn, args, context, info):
        self.fn = fn
        self.args = args
        self.context = context
        self.info = info

    def execute(self, root):
        return self.fn(root, self.args, self.context, self.info)


class Fragment(object):

    def __init__(self, type, field_asts, context=None, info=None):
        self.type = type
        self.field_asts = field_asts
        self.context = context
        self.info = info

    @cached_property
    def partial_resolvers(self):
        return list(get_resolvers(
            self.context,
            self.type,
            self.field_asts
        ))

    @cached_property
    def fragment_container(self):
        try:
            fields = next(zip(*self.partial_resolvers))
        except StopIteration:
            fields = tuple()

        class FragmentInstance(dict):
            # def __init__(self):
                # self.fields = fields
            # _fields = ('c','b','a')
            set = dict.__setitem__
            # def set(self, name, value):
            #     self[name] = value

            def __iter__(self):
                return iter(fields)

        return FragmentInstance

    def have_type(self, root):
        return not self.type.is_type_of or self.type.is_type_of(root, self.context.context_value, self.info)

    def resolve(self, root):
        if root and not self.have_type(root):
            raise GraphQLError(
                u'Expected value of type "{}" but got: {}.'.format(self.type, type(root).__name__),
                self.info.field_asts
            )

        contains_promise = False

        final_results = self.fragment_container()
        # return OrderedDict(
        #     ((field_name, field_resolver(root, field_args, context, info))
        #         for field_name, field_resolver, field_args, context, info in self.partial_resolvers)
        # )
        for response_name, field_resolver in self.partial_resolvers:

            result = field_resolver.execute(root)
            if result is Undefined:
                continue

            if not contains_promise and is_thenable(result):
                contains_promise = True

            final_results[response_name] = result

        if not contains_promise:
            return final_results

        return promise_for_dict(final_results)
        # return {
        #     field_name: field_resolver(root, field_args, context, info)
        #     for field_name, field_resolver, field_args, context, info in self.partial_resolvers
        # }

    def resolve_serially(self, root):
        def execute_field_callback(results, resolver):
            response_name, field_resolver = resolver

            result = field_resolver.execute(root)

            if result is Undefined:
                return results

            if is_thenable(result):
                def collect_result(resolved_result):
                    results[response_name] = resolved_result
                    return results

                return result.then(collect_result)

            results[response_name] = result
            return results

        def execute_field(prev_promise, resolver):
            return prev_promise.then(lambda results: execute_field_callback(results, resolver))

        return functools.reduce(execute_field, self.partial_resolvers, Promise.resolve(self.fragment_container()))

    def __eq__(self, other):
        return isinstance(other, Fragment) and (
            other.type == self.type and
            other.field_asts == self.field_asts and
            other.context == self.context and
            other.info == self.info
        )


class AbstractFragment(object):

    def __init__(self, abstract_type, field_asts, context=None, info=None):
        self.abstract_type = abstract_type
        self.field_asts = field_asts
        self.context = context
        self.info = info
        self._fragments = {}

    @cached_property
    def possible_types(self):
        return self.context.schema.get_possible_types(self.abstract_type)

    @cached_property
    def possible_types_with_is_type_of(self):
        return [
            (type, type.is_type_of) for type in self.possible_types if callable(type.is_type_of)
        ]

    def get_fragment(self, type):
        if isinstance(type, str):
            type = self.context.schema.get_type(type)

        if type not in self._fragments:
            assert type in self.possible_types, (
                'Runtime Object type "{}" is not a possible type for "{}".'
            ).format(type, self.abstract_type)
            self._fragments[type] = Fragment(
                type,
                self.field_asts,
                self.context,
                self.info
            )

        return self._fragments[type]

    def resolve_type(self, result):
        return_type = self.abstract_type
        context = self.context.context_value

        if return_type.resolve_type:
            return return_type.resolve_type(result, context, self.info)

        for type, is_type_of in self.possible_types_with_is_type_of:
            if is_type_of(result, context, self.info):
                return type

    def resolve(self, root):
        _type = self.resolve_type(root)
        fragment = self.get_fragment(_type)
        return fragment.resolve(root)



import base64
import hashlib
import sys
from typing import NewType, Union

from wandb.sdk.lib.paths import StrPath

ETag = NewType("ETag", str)
HexMD5 = NewType("HexMD5", str)
B64MD5 = NewType("B64MD5", str)


def _md5(data: bytes = b"") -> "hashlib._Hash":
    """Allow FIPS-compliant md5 hash when supported."""
    if sys.version_info >= (3, 9):
        return hashlib.md5(data, usedforsecurity=False)
    else:
        return hashlib.md5(data)


def md5_string(string: str) -> B64MD5:
    return _b64_from_hasher(_md5(string.encode("utf-8")))


def _b64_from_hasher(hasher: "hashlib._Hash") -> B64MD5:
    return B64MD5(base64.b64encode(hasher.digest()).decode("ascii"))


def b64_to_hex_id(string: B64MD5) -> HexMD5:
    return HexMD5(base64.standard_b64decode(string).hex())


def hex_to_b64_id(encoded_string: Union[str, bytes]) -> B64MD5:
    if isinstance(encoded_string, bytes):
        encoded_string = encoded_string.decode("utf-8")
    as_str = bytes.fromhex(encoded_string)
    return B64MD5(base64.standard_b64encode(as_str).decode("utf-8"))


def md5_file_b64(*paths: StrPath) -> B64MD5:
    return _b64_from_hasher(_md5_file_hasher(*paths))


def md5_file_hex(*paths: StrPath) -> HexMD5:
    return HexMD5(_md5_file_hasher(*paths).hexdigest())


def _md5_file_hasher(*paths: StrPath) -> "hashlib._Hash":
    md5_hash = _md5()
    for path in sorted(str(p) for p in paths):
        with open(path, "rb") as f:
            for chunk in iter(lambda: f.read(64 * 1024), b""):
                md5_hash.update(chunk)
    return md5_hash



import json
import multiprocessing
import os
import re
import subprocess
import sys
from typing import List, Optional, Set

FAILED_PACKAGES_PREFIX = "ERROR: Failed to install: "
FAILED_PACKAGES_POSTFIX = ". During automated build process."
CORES = multiprocessing.cpu_count()
ONLY_INCLUDE = {x for x in os.getenv("WANDB_ONLY_INCLUDE", "").split(",") if x != ""}
OPTS = []
# If the builder doesn't support buildx no need to use the cache
if os.getenv("WANDB_DISABLE_CACHE"):
    OPTS.append("--no-cache-dir")
# When installing all packages from requirements.frozen.txt no need to resolve deps
if len(ONLY_INCLUDE) == 0:
    OPTS.append("--no-deps")
# When installing the intersection of requirements.frozen.txt and requirements.txt
# force the frozen versions
else:
    OPTS.append("--force")


def install_deps(
    deps: List[str],
    failed: Optional[Set[str]] = None,
    extra_index: Optional[str] = None,
    opts: Optional[List[str]] = None,
) -> Optional[Set[str]]:
    """Install pip dependencies.

    Arguments:
        deps {List[str]} -- List of dependencies to install
        failed (set, None): The libraries that failed to install

    Returns:
        deps (str[], None): The dependencies that failed to install
    """
    try:
        # Include only uri if @ is present
        clean_deps = [d.split("@")[-1].strip() if "@" in d else d for d in deps]
        index_args = ["--extra-index-url", extra_index] if extra_index else []
        print("installing {}...".format(", ".join(clean_deps)))
        opts = opts or []
        args = ["pip", "install"] + opts + clean_deps + index_args
        sys.stdout.flush()
        subprocess.check_output(args, stderr=subprocess.STDOUT)
        return failed
    except subprocess.CalledProcessError as e:
        if failed is None:
            failed = set()
        num_failed = len(failed)
        for line in e.output.decode("utf8").splitlines():
            if line.startswith("ERROR:"):
                clean_dep = find_package_in_error_string(clean_deps, line)
                if clean_dep is not None:
                    if clean_dep in deps:
                        failed.add(clean_dep)
                    else:
                        for d in deps:
                            if clean_dep in d:
                                failed.add(d.replace(" ", ""))
                                break
        if len(set(clean_deps) - failed) == 0:
            return failed
        elif len(failed) > num_failed:
            return install_deps(
                list(set(clean_deps) - failed),
                failed,
                extra_index=extra_index,
                opts=opts,
            )
        else:
            return failed


def main() -> None:
    """Install deps in requirements.frozen.txt."""
    extra_index = None
    torch_reqs = []
    if os.path.exists("requirements.frozen.txt"):
        with open("requirements.frozen.txt") as f:
            print("Installing frozen dependencies...")
            reqs = []
            failed: Set[str] = set()
            for req in f:
                if (
                    len(ONLY_INCLUDE) == 0
                    or req in ONLY_INCLUDE
                    or req.split("=")[0].lower() in ONLY_INCLUDE
                ):
                    # can't pip install wandb==0.*.*.dev1 through pip. Lets just install wandb for now
                    if req.startswith("wandb==") and "dev1" in req:
                        req = "wandb"
                    match = re.match(
                        r"torch(vision|audio)?==\d+\.\d+\.\d+(\+(?:cu[\d]{2,3})|(?:cpu))?",
                        req,
                    )
                    if match:
                        variant = match.group(2)
                        if variant:
                            extra_index = (
                                f"https://download.pytorch.org/whl/{variant[1:]}"
                            )
                        torch_reqs.append(req.strip().replace(" ", ""))
                    else:
                        reqs.append(req.strip().replace(" ", ""))
                else:
                    print(f"Ignoring requirement: {req} from frozen requirements")
                if len(reqs) >= CORES:
                    deps_failed = install_deps(reqs, opts=OPTS)
                    reqs = []
                    if deps_failed is not None:
                        failed = failed.union(deps_failed)
            if len(reqs) > 0:
                deps_failed = install_deps(reqs, opts=OPTS)
                if deps_failed is not None:
                    failed = failed.union(deps_failed)
            with open("_wandb_bootstrap_errors.json", "w") as f:
                f.write(json.dumps({"pip": list(failed)}))
            if len(failed) > 0:
                sys.stderr.write(
                    FAILED_PACKAGES_PREFIX + ",".join(failed) + FAILED_PACKAGES_POSTFIX
                )
                sys.stderr.flush()
        install_deps(torch_reqs, extra_index=extra_index)
    else:
        print("No frozen requirements found")


# hacky way to get the name of the requirement that failed
# attempt last word which is the name of the package often
# fall back to checking all words in the line for the package name
def find_package_in_error_string(deps: List[str], line: str) -> Optional[str]:
    # if the last word in the error string is in the list of deps, return it
    last_word = line.split(" ")[-1]
    if last_word in deps:
        return last_word
    # if the last word is not in the list of deps, check all words
    # TODO: this could report the wrong package if the error string
    # contains a reference to another package in the deps
    # before the package that failed to install
    for word in line.split(" "):
        if word in deps:
            return word
    # if we can't find the package, return None
    return None


if __name__ == "__main__":
    main()



# flake8: noqa
from ._blocks import (
    H1,
    H2,
    H3,
    BlockQuote,
    CalloutBlock,
    CheckedList,
    CodeBlock,
    Gallery,
    HorizontalRule,
    Image,
    LaTeXBlock,
    MarkdownBlock,
    OrderedList,
    P,
    PanelGrid,
    SoundCloud,
    Spotify,
    TableOfContents,
    UnorderedList,
    Video,
    WeaveBlockArtifact,
    WeaveBlockArtifactVersionedFile,
    WeaveBlockSummaryTable,
)



#
import json
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Union

from wandb.proto import wandb_internal_pb2 as pb

if TYPE_CHECKING:  # pragma: no cover
    from google.protobuf.internal.containers import (
        MessageMap,
        RepeatedCompositeFieldContainer,
    )
    from google.protobuf.message import Message

    from wandb.proto import wandb_server_pb2 as spb
    from wandb.proto import wandb_telemetry_pb2 as tpb


def dict_from_proto_list(obj_list: "RepeatedCompositeFieldContainer") -> Dict[str, Any]:
    return {item.key: json.loads(item.value_json) for item in obj_list}


def _result_from_record(record: "pb.Record") -> "pb.Result":
    result = pb.Result(uuid=record.uuid, control=record.control)
    return result


def _assign_record_num(record: "pb.Record", record_num: int) -> None:
    record.num = record_num


def _assign_end_offset(record: "pb.Record", end_offset: int) -> None:
    record.control.end_offset = end_offset


def proto_encode_to_dict(
    pb_obj: Union["tpb.TelemetryRecord", "pb.MetricRecord"]
) -> Dict[int, Any]:
    data: Dict[int, Any] = dict()
    fields = pb_obj.ListFields()
    for desc, value in fields:
        if desc.name.startswith("_"):
            continue
        if desc.type == desc.TYPE_STRING:
            data[desc.number] = value
        elif desc.type == desc.TYPE_INT32:
            data[desc.number] = value
        elif desc.type == desc.TYPE_ENUM:
            data[desc.number] = value
        elif desc.type == desc.TYPE_MESSAGE:
            nested = value.ListFields()
            bool_msg = all(d.type == d.TYPE_BOOL for d, _ in nested)
            if bool_msg:
                items = [d.number for d, v in nested if v]
                if items:
                    data[desc.number] = items
            else:
                # TODO: for now this code only handles sub-messages with strings
                md = {}
                for d, v in nested:
                    if not v or d.type != d.TYPE_STRING:
                        continue
                    md[d.number] = v
                data[desc.number] = md
    return data


def settings_dict_from_pbmap(
    pbmap: "MessageMap[str, spb.SettingsValue]",
) -> Dict[str, Any]:
    d: Dict[str, Any] = dict()
    for k in pbmap:
        v_obj = pbmap[k]
        v_type = v_obj.WhichOneof("value_type")

        v: Union[int, str, float, None, tuple, dict, datetime] = None
        if v_type == "int_value":
            v = v_obj.int_value
        elif v_type == "string_value":
            v = v_obj.string_value
        elif v_type == "float_value":
            v = v_obj.float_value
        elif v_type == "bool_value":
            v = v_obj.bool_value
        elif v_type == "null_value":
            v = None
        elif v_type == "tuple_value":
            v = tuple(v_obj.tuple_value.string_values)
        elif v_type == "map_value":
            v = dict(v_obj.map_value.map_values)
        elif v_type == "nested_map_value":
            v = {
                k: dict(vv.map_values)
                for k, vv in dict(v_obj.nested_map_value.nested_map_values).items()
            }
        elif v_type == "timestamp_value":
            v = datetime.strptime(v_obj.timestamp_value, "%Y%m%d_%H%M%S")
        d[k] = v
    return d


def message_to_dict(
    message: "Message",
) -> Dict[str, Any]:
    """Convert a protobuf message into a dictionary."""
    from google.protobuf.json_format import MessageToDict

    return MessageToDict(message, preserving_proto_field_name=True)



import wandb
from wandb import util
from wandb.plots.utils import (
    deprecation_notice,
    encode_labels,
    test_missing,
    test_types,
)

chart_limit = wandb.Table.MAX_ROWS


def roc(
    y_true=None,
    y_probas=None,
    labels=None,
    plot_micro=True,
    plot_macro=True,
    classes_to_plot=None,
):
    """
    Calculates receiver operating characteristic scores and visualizes them as the
     ROC curve.

    Arguments:
     y_true (arr): Test set labels.
     y_probas (arr): Test set predicted probabilities.
     labels (list): Named labels for target varible (y). Makes plots easier to
                     read by replacing target values with corresponding index.
                     For example labels= ['dog', 'cat', 'owl'] all 0s are
                     replaced by 'dog', 1s by 'cat'.

    Returns:
     Nothing. To see plots, go to your W&B run page then expand the 'media' tab
           under 'auto visualizations'.

    Example:
     wandb.log({'roc': wandb.plots.ROC(y_true, y_probas, labels)})
    """
    deprecation_notice()

    np = util.get_module(
        "numpy",
        required="roc requires the numpy library, install with `pip install numpy`",
    )
    sklearn = util.get_module(
        "sklearn",
        required="roc requires the scikit library, install with `pip install scikit-learn`",
    )
    from sklearn.metrics import auc, roc_curve

    if test_missing(y_true=y_true, y_probas=y_probas) and test_types(
        y_true=y_true, y_probas=y_probas
    ):
        y_true = np.array(y_true)
        y_probas = np.array(y_probas)
        classes = np.unique(y_true)
        probas = y_probas

        if classes_to_plot is None:
            classes_to_plot = classes

        fpr_dict = dict()
        tpr_dict = dict()

        indices_to_plot = np.in1d(classes, classes_to_plot)

        def roc_table(fpr_dict, tpr_dict, classes, indices_to_plot):
            data = []
            count = 0

            for i, to_plot in enumerate(indices_to_plot):
                fpr_dict[i], tpr_dict[i], _ = roc_curve(
                    y_true, probas[:, i], pos_label=classes[i]
                )
                if to_plot:
                    roc_auc = auc(fpr_dict[i], tpr_dict[i])
                    for j in range(len(fpr_dict[i])):
                        if labels is not None and (
                            isinstance(classes[i], int)
                            or isinstance(classes[0], np.integer)
                        ):
                            class_dict = labels[classes[i]]
                        else:
                            class_dict = classes[i]
                        fpr = [
                            class_dict,
                            round(fpr_dict[i][j], 3),
                            round(tpr_dict[i][j], 3),
                        ]
                        data.append(fpr)
                        count += 1
                        if count >= chart_limit:
                            wandb.termwarn(
                                "wandb uses only the first %d datapoints to create the plots."
                                % wandb.Table.MAX_ROWS
                            )
                            break
            return wandb.visualize(
                "wandb/roc/v1", wandb.Table(columns=["class", "fpr", "tpr"], data=data)
            )

        return roc_table(fpr_dict, tpr_dict, classes, indices_to_plot)



"""Router - handle message router (base class).

Router to manage responses.

"""

import logging
import threading
import uuid
from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, Optional

from ..lib import mailbox, tracelog
from .message_future import MessageFuture

if TYPE_CHECKING:
    from queue import Queue

    from wandb.proto import wandb_internal_pb2 as pb


logger = logging.getLogger("wandb")


class MessageRouterClosedError(Exception):
    """Router has been closed."""

    pass


class MessageFutureObject(MessageFuture):
    def __init__(self) -> None:
        super().__init__()

    def get(self, timeout: Optional[int] = None) -> Optional["pb.Result"]:
        is_set = self._object_ready.wait(timeout)
        if is_set and self._object:
            return self._object
        return None


class MessageRouter:
    _pending_reqs: Dict[str, MessageFutureObject]
    _request_queue: "Queue[pb.Record]"
    _response_queue: "Queue[pb.Result]"
    _mailbox: Optional[mailbox.Mailbox]

    def __init__(self, mailbox: Optional[mailbox.Mailbox] = None) -> None:
        self._mailbox = mailbox
        self._pending_reqs = {}
        self._lock = threading.Lock()

        self._join_event = threading.Event()
        self._thread = threading.Thread(target=self.message_loop)
        self._thread.name = "MsgRouterThr"
        self._thread.daemon = True
        self._thread.start()

    @abstractmethod
    def _read_message(self) -> Optional["pb.Result"]:
        raise NotImplementedError

    @abstractmethod
    def _send_message(self, record: "pb.Record") -> None:
        raise NotImplementedError

    def message_loop(self) -> None:
        while not self._join_event.is_set():
            try:
                msg = self._read_message()
            except EOFError:
                # On abnormal shutdown the queue will be destroyed underneath
                # resulting in EOFError.  message_loop needs to exit..
                logger.warning("EOFError seen in message_loop")
                break
            except MessageRouterClosedError:
                logger.warning("message_loop has been closed")
                break
            if not msg:
                continue
            self._handle_msg_rcv(msg)

    def send_and_receive(
        self, rec: "pb.Record", local: Optional[bool] = None
    ) -> MessageFuture:
        rec.control.req_resp = True
        if local:
            rec.control.local = local
        rec.uuid = uuid.uuid4().hex
        future = MessageFutureObject()
        with self._lock:
            self._pending_reqs[rec.uuid] = future

        self._send_message(rec)

        return future

    def join(self) -> None:
        self._join_event.set()
        self._thread.join()

    def _handle_msg_rcv(self, msg: "pb.Result") -> None:
        # deliver mailbox addressed messages to mailbox
        if self._mailbox and msg.control.mailbox_slot:
            self._mailbox.deliver(msg)
            return
        with self._lock:
            future = self._pending_reqs.pop(msg.uuid, None)
        if future is None:
            # TODO (cvp): saw this in tests, seemed benign enough to ignore, but
            # could point to other issues.
            if msg.uuid != "":
                tracelog.log_message_assert(msg)
                logger.warning(
                    "No listener found for msg with uuid %s (%s)", msg.uuid, msg
                )
            return
        future._set_object(msg)



#!/usr/bin/env python
"""Finite state machine.

Simple FSM implementation.

Usage:
    ```python
    class A:
        def on_output(self, inputs) -> None:
            pass

    class B:
        def on_output(self, inputs) -> None:
            pass

    def to_b(inputs) -> bool:
        return True

    def to_a(inputs) -> bool:
        return True

    f = Fsm(states=[A(), B()],
            table={A: [(to_b, B)],
                   B: [(to_a, A)]})
    f.run({"input1": 1, "input2": 2})
    ```
"""

import sys
from abc import abstractmethod
from dataclasses import dataclass
from typing import Callable, Dict, Generic, Optional, Sequence, Type, TypeVar, Union

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

T_FsmInputs = TypeVar("T_FsmInputs", contravariant=True)
T_FsmContext = TypeVar("T_FsmContext")
T_FsmContext_cov = TypeVar("T_FsmContext_cov", covariant=True)
T_FsmContext_contra = TypeVar("T_FsmContext_contra", contravariant=True)


@runtime_checkable
class FsmStateCheck(Protocol[T_FsmInputs]):
    @abstractmethod
    def on_check(self, inputs: T_FsmInputs) -> None:
        ...  # pragma: no cover


@runtime_checkable
class FsmStateOutput(Protocol[T_FsmInputs]):
    @abstractmethod
    def on_state(self, inputs: T_FsmInputs) -> None:
        ...  # pragma: no cover


@runtime_checkable
class FsmStateEnter(Protocol[T_FsmInputs]):
    @abstractmethod
    def on_enter(self, inputs: T_FsmInputs) -> None:
        ...  # pragma: no cover


@runtime_checkable
class FsmStateEnterWithContext(Protocol[T_FsmInputs, T_FsmContext_contra]):
    @abstractmethod
    def on_enter(self, inputs: T_FsmInputs, context: T_FsmContext_contra) -> None:
        ...  # pragma: no cover


@runtime_checkable
class FsmStateStay(Protocol[T_FsmInputs]):
    @abstractmethod
    def on_stay(self, inputs: T_FsmInputs) -> None:
        ...  # pragma: no cover


@runtime_checkable
class FsmStateExit(Protocol[T_FsmInputs, T_FsmContext_cov]):
    @abstractmethod
    def on_exit(self, inputs: T_FsmInputs) -> T_FsmContext_cov:
        ...  # pragma: no cover


# It would be nice if python provided optional protocol members, but it doesnt as described here:
#   https://peps.python.org/pep-0544/#support-optional-protocol-members
# Until then, we can only enforce that a state at least supports one protocol interface.  This
# unfortunately will not check the signature of other potential protocols.
FsmState: TypeAlias = Union[
    FsmStateCheck[T_FsmInputs],
    FsmStateOutput[T_FsmInputs],
    FsmStateEnter[T_FsmInputs],
    FsmStateEnterWithContext[T_FsmInputs, T_FsmContext],
    FsmStateStay[T_FsmInputs],
    FsmStateExit[T_FsmInputs, T_FsmContext],
]


@dataclass
class FsmEntry(Generic[T_FsmInputs, T_FsmContext]):
    condition: Callable[[T_FsmInputs], bool]
    target_state: Type[FsmState[T_FsmInputs, T_FsmContext]]
    action: Optional[Callable[[T_FsmInputs], None]] = None


FsmTableWithContext: TypeAlias = Dict[
    Type[FsmState[T_FsmInputs, T_FsmContext]],
    Sequence[FsmEntry[T_FsmInputs, T_FsmContext]],
]


FsmTable: TypeAlias = FsmTableWithContext[T_FsmInputs, None]


class FsmWithContext(Generic[T_FsmInputs, T_FsmContext]):
    _state_dict: Dict[Type[FsmState], FsmState]
    _table: FsmTableWithContext[T_FsmInputs, T_FsmContext]
    _state: FsmState[T_FsmInputs, T_FsmContext]
    _states: Sequence[FsmState]

    def __init__(
        self,
        states: Sequence[FsmState],
        table: FsmTableWithContext[T_FsmInputs, T_FsmContext],
    ) -> None:
        self._states = states
        self._table = table
        self._state_dict = {type(s): s for s in states}
        self._state = self._state_dict[type(states[0])]

    def _transition(
        self,
        inputs: T_FsmInputs,
        new_state: Type[FsmState[T_FsmInputs, T_FsmContext]],
        action: Optional[Callable[[T_FsmInputs], None]],
    ) -> None:
        if action:
            action(inputs)

        context = None
        if isinstance(self._state, FsmStateExit):
            context = self._state.on_exit(inputs)

        prev_state = type(self._state)
        if prev_state == new_state:
            if isinstance(self._state, FsmStateStay):
                self._state.on_stay(inputs)
        else:
            self._state = self._state_dict[new_state]
            if context and isinstance(self._state, FsmStateEnterWithContext):
                self._state.on_enter(inputs, context=context)
            elif isinstance(self._state, FsmStateEnter):
                self._state.on_enter(inputs)

    def _check_transitions(self, inputs: T_FsmInputs) -> None:
        for entry in self._table[type(self._state)]:
            if entry.condition(inputs):
                self._transition(inputs, entry.target_state, entry.action)
                return

    def input(self, inputs: T_FsmInputs) -> None:
        if isinstance(self._state, FsmStateCheck):
            self._state.on_check(inputs)
        self._check_transitions(inputs)
        if isinstance(self._state, FsmStateOutput):
            self._state.on_state(inputs)


Fsm: TypeAlias = FsmWithContext[T_FsmInputs, None]



from collections.abc import Iterable
import decimal
from functools import partial

from wandb_graphql.language import ast
from wandb_graphql.language.printer import print_ast
from wandb_graphql.type import (GraphQLField, GraphQLList,
                          GraphQLNonNull, GraphQLEnumType)

from .utils import to_camel_case


class DSLSchema(object):
    def __init__(self, client):
        self.client = client

    @property
    def schema(self):
        return self.client.schema

    def __getattr__(self, name):
        type_def = self.schema.get_type(name)
        return DSLType(type_def)

    def query(self, *args, **kwargs):
        return self.execute(query(*args, **kwargs))

    def mutate(self, *args, **kwargs):
        return self.query(*args, operation='mutate', **kwargs)

    def execute(self, document):
        return self.client.execute(document)


class DSLType(object):
    def __init__(self, type):
        self.type = type

    def __getattr__(self, name):
        formatted_name, field_def = self.get_field(name)
        return DSLField(formatted_name, field_def)

    def get_field(self, name):
        camel_cased_name = to_camel_case(name)

        if name in self.type.fields:
            return name, self.type.fields[name]

        if camel_cased_name in self.type.fields:
            return camel_cased_name, self.type.fields[camel_cased_name]

        raise KeyError('Field {} doesnt exist in type {}.'.format(name, self.type.name))


def selections(*fields):
    for _field in fields:
        yield field(_field).ast


def get_ast_value(value):
    if isinstance(value, ast.Node):
        return value
    if isinstance(value, str):
        return ast.StringValue(value=value)
    elif isinstance(value, bool):
        return ast.BooleanValue(value=value)
    elif isinstance(value, (float, decimal.Decimal)):
        return ast.FloatValue(value=value)
    elif isinstance(value, int):
        return ast.IntValue(value=value)
    return None


class DSLField(object):

    def __init__(self, name, field):
        self.field = field
        self.ast_field = ast.Field(name=ast.Name(value=name), arguments=[])
        self.selection_set = None

    def select(self, *fields):
        if not self.ast_field.selection_set:
            self.ast_field.selection_set = ast.SelectionSet(selections=[])
        self.ast_field.selection_set.selections.extend(selections(*fields))
        return self

    def __call__(self, *args, **kwargs):
        return self.args(*args, **kwargs)

    def alias(self, alias):
        self.ast_field.alias = ast.Name(value=alias)
        return self

    def args(self, **args):
        for name, value in args.items():
            arg = self.field.args.get(name)
            arg_type_serializer = get_arg_serializer(arg.type)
            value = arg_type_serializer(value)
            self.ast_field.arguments.append(
                ast.Argument(
                    name=ast.Name(value=name),
                    value=get_ast_value(value)
                )
            )
        return self

    @property
    def ast(self):
        return self.ast_field

    def __str__(self):
        return print_ast(self.ast_field)


def field(field, **args):
    if isinstance(field, GraphQLField):
        return DSLField(field).args(**args)
    elif isinstance(field, DSLField):
        return field

    raise Exception('Received incompatible query field: "{}".'.format(field))


def query(*fields):
    return ast.Document(
        definitions=[ast.OperationDefinition(
            operation='query',
            selection_set=ast.SelectionSet(
                selections=list(selections(*fields))
            )
        )]
    )


def serialize_list(serializer, values):
    assert isinstance(values, Iterable), 'Expected iterable, received "{}"'.format(repr(values))
    return [serializer(v) for v in values]


def get_arg_serializer(arg_type):
    if isinstance(arg_type, GraphQLNonNull):
        return get_arg_serializer(arg_type.of_type)
    if isinstance(arg_type, GraphQLList):
        inner_serializer = get_arg_serializer(arg_type.of_type)
        return partial(serialize_list, inner_serializer)
    if isinstance(arg_type, GraphQLEnumType):
        return lambda value: ast.EnumValue(value=arg_type.serialize(value))
    return arg_type.serialize


def var(name):
    return ast.Variable(name=name)



"""leveldb log datastore.

Format is described at:
    https://github.com/google/leveldb/blob/master/doc/log_format.md

block := record* trailer?
record :=
  checksum: uint32     // crc32c of type and data[] ; little-endian
  length: uint16       // little-endian
  type: uint8          // One of FULL, FIRST, MIDDLE, LAST
  data: uint8[length]

header :=
  ident: char[4]
  magic: uint16
  version: uint8
"""

# TODO: possibly restructure code by porting the C++ or go implementation

import logging
import os
import struct
import zlib
from typing import TYPE_CHECKING, Tuple

import wandb

if TYPE_CHECKING:
    from wandb.proto.wandb_internal_pb2 import Record

logger = logging.getLogger(__name__)

LEVELDBLOG_HEADER_LEN = 7
LEVELDBLOG_BLOCK_LEN = 32768
LEVELDBLOG_DATA_LEN = LEVELDBLOG_BLOCK_LEN - LEVELDBLOG_HEADER_LEN

LEVELDBLOG_FULL = 1
LEVELDBLOG_FIRST = 2
LEVELDBLOG_MIDDLE = 3
LEVELDBLOG_LAST = 4

LEVELDBLOG_HEADER_IDENT = ":W&B"
LEVELDBLOG_HEADER_MAGIC = (
    0xBEE1  # zlib.crc32(bytes("Weights & Biases", 'iso8859-1')) & 0xffff
)
LEVELDBLOG_HEADER_VERSION = 0

try:
    bytes("", "ascii")

    def strtobytes(x):
        """strtobytes."""
        return bytes(x, "iso8859-1")

    # def bytestostr(x):
    #     return str(x, 'iso8859-1')

except Exception:
    strtobytes = str
    # bytestostr = str


class DataStore:
    _index: int
    _flush_offset: int

    def __init__(self) -> None:
        self._opened_for_scan = False
        self._fp = None
        self._index = 0
        self._flush_offset = 0
        self._size_bytes = 0

        self._crc = [0] * (LEVELDBLOG_LAST + 1)
        for x in range(1, LEVELDBLOG_LAST + 1):
            self._crc[x] = zlib.crc32(strtobytes(chr(x))) & 0xFFFFFFFF

        assert (
            wandb._assert_is_internal_process
        ), "DataStore can only be used in the internal process"

    def open_for_write(self, fname: str) -> None:
        self._fname = fname
        logger.info("open: %s", fname)
        open_flags = "xb"
        self._fp = open(fname, open_flags)
        self._write_header()

    def open_for_append(self, fname):
        # TODO: implement
        self._fname = fname
        logger.info("open: %s", fname)
        self._fp = open(fname, "wb")
        # do something with _index

    def open_for_scan(self, fname):
        self._fname = fname
        logger.info("open for scan: %s", fname)
        self._fp = open(fname, "r+b")
        self._index = 0
        self._size_bytes = os.stat(fname).st_size
        self._opened_for_scan = True
        self._read_header()

    def seek(self, offset: int) -> None:
        self._fp.seek(offset)
        self._index = offset

    def get_offset(self) -> int:
        offset = self._fp.tell()
        return offset

    def in_last_block(self):
        """Determine if we're in the last block to handle in-progress writes."""
        return self._index > self._size_bytes - LEVELDBLOG_DATA_LEN

    def scan_record(self):
        assert self._opened_for_scan, "file not open for scanning"
        # TODO(jhr): handle some assertions as file corruption issues
        # assume we have enough room to read header, checked by caller?
        header = self._fp.read(LEVELDBLOG_HEADER_LEN)
        if len(header) == 0:
            return None
        assert (
            len(header) == LEVELDBLOG_HEADER_LEN
        ), "record header is {} bytes instead of the expected {}".format(
            len(header), LEVELDBLOG_HEADER_LEN
        )
        fields = struct.unpack("<IHB", header)
        checksum, dlength, dtype = fields
        # check len, better fit in the block
        self._index += LEVELDBLOG_HEADER_LEN
        data = self._fp.read(dlength)
        checksum_computed = zlib.crc32(data, self._crc[dtype]) & 0xFFFFFFFF
        assert (
            checksum == checksum_computed
        ), "record checksum is invalid, data may be corrupt"
        self._index += dlength
        return dtype, data

    def scan_data(self):
        # TODO(jhr): handle some assertions as file corruption issues
        # how much left in the block.  if less than header len, read as pad,
        offset = self._index % LEVELDBLOG_BLOCK_LEN
        space_left = LEVELDBLOG_BLOCK_LEN - offset
        if space_left < LEVELDBLOG_HEADER_LEN:
            pad_check = strtobytes("\x00" * space_left)
            pad = self._fp.read(space_left)
            # verify they are zero
            assert pad == pad_check, "invalid padding"
            self._index += space_left

        record = self.scan_record()
        if record is None:  # eof
            return None
        dtype, data = record
        if dtype == LEVELDBLOG_FULL:
            return data

        assert (
            dtype == LEVELDBLOG_FIRST
        ), f"expected record to be type {LEVELDBLOG_FIRST} but found {dtype}"
        while True:
            offset = self._index % LEVELDBLOG_BLOCK_LEN
            record = self.scan_record()
            if record is None:  # eof
                return None
            dtype, new_data = record
            if dtype == LEVELDBLOG_LAST:
                data += new_data
                break
            assert (
                dtype == LEVELDBLOG_MIDDLE
            ), "expected record to be type {} but found {}".format(
                LEVELDBLOG_MIDDLE, dtype
            )
            data += new_data
        return data

    def _write_header(self):
        data = struct.pack(
            "<4sHB",
            strtobytes(LEVELDBLOG_HEADER_IDENT),
            LEVELDBLOG_HEADER_MAGIC,
            LEVELDBLOG_HEADER_VERSION,
        )
        assert (
            len(data) == LEVELDBLOG_HEADER_LEN
        ), "header size is {} bytes, expected {}".format(
            len(data), LEVELDBLOG_HEADER_LEN
        )
        self._fp.write(data)
        self._index += len(data)

    def _read_header(self):
        header = self._fp.read(LEVELDBLOG_HEADER_LEN)
        assert (
            len(header) == LEVELDBLOG_HEADER_LEN
        ), "header is {} bytes instead of the expected {}".format(
            len(header), LEVELDBLOG_HEADER_LEN
        )
        ident, magic, version = struct.unpack("<4sHB", header)
        if ident != strtobytes(LEVELDBLOG_HEADER_IDENT):
            raise Exception("Invalid header")
        if magic != LEVELDBLOG_HEADER_MAGIC:
            raise Exception("Invalid header")
        if version != LEVELDBLOG_HEADER_VERSION:
            raise Exception("Invalid header")
        self._index += len(header)

    def _write_record(self, s, dtype=None):
        """Write record that must fit into a block."""
        # double check that there is enough space
        # (this is a precondition to calling this method)
        assert len(s) + LEVELDBLOG_HEADER_LEN <= (
            LEVELDBLOG_BLOCK_LEN - self._index % LEVELDBLOG_BLOCK_LEN
        ), "not enough space to write new records"

        dlength = len(s)
        dtype = dtype or LEVELDBLOG_FULL
        # print("record: length={} type={}".format(dlength, dtype))
        checksum = zlib.crc32(s, self._crc[dtype]) & 0xFFFFFFFF
        # logger.info("write_record: index=%d len=%d dtype=%d",
        #     self._index, dlength, dtype)
        self._fp.write(struct.pack("<IHB", checksum, dlength, dtype))
        if dlength:
            self._fp.write(s)
        self._index += LEVELDBLOG_HEADER_LEN + len(s)

    def _write_data(self, s):
        start_offset = self._index

        offset = self._index % LEVELDBLOG_BLOCK_LEN
        space_left = LEVELDBLOG_BLOCK_LEN - offset
        data_used = 0
        data_left = len(s)
        # logger.info("write_data: index=%d offset=%d len=%d",
        #     self._index, offset, data_left)
        if space_left < LEVELDBLOG_HEADER_LEN:
            pad = "\x00" * space_left
            self._fp.write(strtobytes(pad))
            self._index += space_left
            offset = 0
            space_left = LEVELDBLOG_BLOCK_LEN

        # does it fit in first (possibly partial) block?
        if data_left + LEVELDBLOG_HEADER_LEN <= space_left:
            self._write_record(s)
        else:
            # write first record (we could still be in the middle of a block,
            # but this write will end on a block boundary)
            data_room = space_left - LEVELDBLOG_HEADER_LEN
            self._write_record(s[:data_room], LEVELDBLOG_FIRST)
            data_used += data_room
            data_left -= data_room
            assert data_left, "data_left should be non-zero"

            # write middles (if any)
            while data_left > LEVELDBLOG_DATA_LEN:
                self._write_record(
                    s[data_used : data_used + LEVELDBLOG_DATA_LEN],
                    LEVELDBLOG_MIDDLE,
                )
                data_used += LEVELDBLOG_DATA_LEN
                data_left -= LEVELDBLOG_DATA_LEN

            # write last and flush the entire block to disk
            self._write_record(s[data_used:], LEVELDBLOG_LAST)
            self._fp.flush()
            os.fsync(self._fp.fileno())
            self._flush_offset = self._index

        return start_offset, self._index, self._flush_offset

    def ensure_flushed(self, off: int) -> None:
        self._fp.flush()

    def write(self, obj: "Record") -> Tuple[int, int, int]:
        """Write a protocol buffer.

        Arguments:
            obj: Protocol buffer to write.

        Returns:
            (start_offset, end_offset, flush_offset) if successful,
            None otherwise

        """
        raw_size = obj.ByteSize()
        s = obj.SerializeToString()
        assert len(s) == raw_size, "invalid serialization"
        ret = self._write_data(s)
        return ret

    def close(self) -> None:
        if self._fp is not None:
            logger.info("close: %s", self._fname)
            self._fp.close()



# We can import from the top after dropping support for Python 3.6
# import wandb.apis.reports as wr
from .util import coalesce


def create_example_header():
    """Create an example header with image at top."""
    import wandb.apis.reports as wr

    return [
        wr.P(),
        wr.HorizontalRule(),
        wr.P(),
        wr.Image(
            "https://camo.githubusercontent.com/83839f20c90facc062330f8fee5a7ab910fdd04b80b4c4c7e89d6d8137543540/68747470733a2f2f692e696d6775722e636f6d2f676236423469672e706e67"
        ),
        wr.P(),
        wr.HorizontalRule(),
        wr.P(),
    ]


def create_example_footer():
    """Create an example footer with image and text at bottom."""
    import wandb.apis.reports as wr

    return [
        wr.P(),
        wr.HorizontalRule(),
        wr.P(),
        wr.H1("Disclaimer"),
        wr.P(
            "The views and opinions expressed in this report are those of the authors and do not necessarily reflect the official policy or position of Weights & Biases. blah blah blah blah blah boring text at the bottom"
        ),
        wr.P(),
        wr.HorizontalRule(),
    ]


def create_enterprise_report(
    project=None,
    title="Untitled Report",
    description="",
    header=None,
    body=None,
    footer=None,
):
    """Create an example enterprise report with a header and footer.

    Can be used to add custom branding to reports.
    """
    import wandb.apis.reports as wr

    project = coalesce(project, "default-project")
    header = coalesce(header, create_example_header())
    body = coalesce(body, [])
    footer = coalesce(footer, create_example_footer())

    return wr.Report(
        project=project,
        title=title,
        description=description,
        blocks=[*header, *body, *footer],
    )


def create_customer_landing_page(
    project=None,
    company_name="My Company",
    main_contact="My Contact (name@email.com)",
    slack_link="https://company.slack.com",
):
    """Create an example customer landing page using data from Andrew's demo."""
    import wandb.apis.reports as wr

    project = coalesce(project, "default-project")

    return wr.Report(
        project,
        title=f"Weights & Biases @ {company_name}",
        description=f"The developer-first MLOps platform is now available at {company_name}!\nReach out to {main_contact} for an account, and join your dedicated slack channel at:\n{slack_link}",
        blocks=[
            wr.P(),
            wr.HorizontalRule(),
            wr.TableOfContents(),
            wr.P(),
            wr.HorizontalRule(),
            wr.H1(text=["What is Weights & Biases?"]),
            wr.P(
                text=[
                    "Weights & Biases (W&B) is the developer-first MLOps platform to build better models faster.  Over 200,000+ ML practitioners at 500+ companies use W&B to optimize their ML workflows in Natural Language, Computer Vision, Reinforcement Learning, Tabular ML, Finance, and more!"
                ]
            ),
            wr.P(),
            wr.H2(text=["Why do you need W&B?"]),
            wr.P(
                text=[
                    "ML is a highly experimental field.  Often we try many different datasets, model architectures, optimizers, hyperparameters, etc."
                ]
            ),
            wr.P(
                text=["Experimentation is great, but it can get messy.  Have you ever:"]
            ),
            wr.UnorderedList(
                items=[
                    ["Logged experiments in a sketchy spreadsheet?"],
                    [
                        "Built an amazing model but could not reproduce it for a colleague / model validation?"
                    ],
                    ["Wondered why your model is making strange predictions?"],
                    ["Fumbled with tuning hyperparameters?"],
                    [
                        "Struggled explaining to a colleague the impact of what you're doing?"
                    ],
                ]
            ),
            wr.P(
                text=["If that sounds familiar, W&B might be a good solution for you!"]
            ),
            wr.P(),
            wr.H2(
                text=[
                    "What does W&B do?",
                    wr.Link(text="", url="https://wandb.ai/site/experiment-tracking"),
                ]
            ),
            wr.P(
                text=[
                    wr.Link(text="", url="https://wandb.ai/site/experiment-tracking"),
                    "W&B has lightweight and flexible tools for... (expand to see more)",
                ]
            ),
            wr.H3(
                text=[
                    wr.Link(
                        text="Experiment tracking",
                        url="https://wandb.ai/site/experiment-tracking",
                    )
                ]
            ),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(
                        entity="megatruong",
                        project="whirlwind_test4",
                        name="Run set",
                        query="",
                        filters={
                            "$or": [
                                {
                                    "$and": [
                                        {"state": {"$ne": "crashed"}},
                                        {
                                            "config.Learner.value.opt_func": {
                                                "$ne": None
                                            }
                                        },
                                    ]
                                }
                            ]
                        },
                        groupby=["Learner.opt_func"],
                        order=["-CreatedTimestamp"],
                    )
                ],
                panels=[
                    wr.LinePlot(
                        x="Step",
                        y=["gradients/layers.0.4.0.bn1.bias"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 16, "y": 12, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["gradients/layers.0.1.weight"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 8, "y": 12, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["gradients/layers.0.1.bias"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 0, "y": 12, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["train_loss"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 16, "y": 0, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["valid_loss"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 16, "y": 6, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["top_k_accuracy"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 8, "y": 0, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["mom_0"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 0, "y": 6, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["lr_0"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 8, "y": 6, "w": 8, "h": 6},
                    ),
                    wr.LinePlot(
                        x="Step",
                        y=["accuracy"],
                        log_y=False,
                        groupby="None",
                        layout={"x": 0, "y": 0, "w": 8, "h": 6},
                    ),
                ],
                custom_run_colors={
                    ("Run set", "megatruong"): "rgb(83, 135, 221)",
                    ("Run set", "fastai.optimizer.ranger"): "rgb(83, 135, 221)",
                    ("Run set", "fastai.optimizer.Adam"): "rgb(229, 116, 57)",
                },
            ),
            wr.P(
                text=[
                    wr.Link(
                        text="",
                        url="https://assets.website-files.com/5ac6b7f2924c656f2b13a88c/6066c22135b8983b61ad7939_weights-and-biases-logo.svg",
                    )
                ]
            ),
            wr.H3(
                text=[
                    wr.Link(
                        text="Dataset and model versioning, evaluation, and reproduction",
                        url="https://wandb.ai/site/artifacts",
                    )
                ]
            ),
            wr.WeaveBlockArtifact(
                entity="megatruong",
                project="whirlwind_test4",
                artifact="camvid_learner",
                tab="lineage",
            ),
            wr.P(),
            wr.H3(
                text=[
                    wr.Link(
                        text="Hyperparameter optimization",
                        url="https://wandb.ai/site/sweeps",
                    )
                ]
            ),
            wr.P(text=[wr.Link(text="", url="https://wandb.ai/site/sweeps")]),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(
                        entity="wandb",
                        project="cartpole",
                        name="Run set",
                        query="sweep",
                        filters={"$or": [{"$and": []}]},
                        order=["-CreatedTimestamp"],
                    )
                ],
                panels=[
                    wr.MediaBrowser(layout={"x": 0, "y": 10, "w": 24, "h": 10}),
                    wr.ParallelCoordinatesPlot(
                        columns=[
                            wr.PCColumn(metric="c::activation"),
                            wr.PCColumn(metric="c::lr", log_scale=True),
                            wr.PCColumn(
                                metric="c::target_model_update",
                                log_scale=True,
                            ),
                            wr.PCColumn(metric="c::n_hidden", log_scale=True),
                            wr.PCColumn(metric="test_reward"),
                        ],
                        layout={"x": 0, "y": 0, "w": 24, "h": 10},
                    ),
                ],
            ),
            wr.H3(
                text=[
                    wr.Link(
                        text="Model visualization and analysis",
                        url="https://wandb.ai/site/tables",
                    )
                ]
            ),
            wr.P(text=[wr.Link(text="", url="https://wandb.ai/site/tables")]),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(
                        entity="megatruong",
                        project="whirlwind_test4",
                        name="Run set",
                        query="",
                        filters={"$or": [{"$and": []}]},
                        order=["-CreatedTimestamp"],
                    )
                ],
                panels=[
                    wr.WeavePanelSummaryTable(
                        table_name="valid_table",
                        layout={"x": 7, "y": 0, "w": 7, "h": 13},
                    ),
                    wr.WeavePanelSummaryTable(
                        table_name="img_table",
                        layout={"x": 0, "y": 0, "w": 7, "h": 13},
                    ),
                    wr.WeavePanelSummaryTable(
                        table_name="image_table",
                        layout={"x": 14, "y": 0, "w": 10, "h": 13},
                    ),
                ],
            ),
            wr.P(),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(
                        entity="wandb",
                        project="wandb_spacy_integration",
                        name="Run set",
                        query="",
                        filters={"$or": [{"$and": []}]},
                        order=["-CreatedTimestamp"],
                    )
                ],
                panels=[
                    wr.WeavePanelSummaryTable(
                        table_name="spaCy NER table",
                        layout={"x": 0, "y": 0, "w": 24, "h": 10},
                    ),
                    wr.WeavePanelSummaryTable(
                        table_name="per annotation scores",
                        layout={"x": 7, "y": 10, "w": 17, "h": 8},
                    ),
                    wr.WeavePanelSummaryTable(
                        table_name="metrics", layout={"x": 0, "y": 10, "w": 7, "h": 8}
                    ),
                ],
            ),
            wr.H3(
                text=[
                    wr.Link(
                        text="ML team collaboration and sharing results",
                        url="https://wandb.ai/site/reports",
                    )
                ]
            ),
            wr.H2(text=["How do I get access?"]),
            wr.P(text=[f"Ask {main_contact} to help:"]),
            wr.OrderedList(
                items=[
                    ["Set up your account"],
                    [
                        "Get added to the ",
                        wr.Link(
                            text="joint slack channel",
                            url=slack_link,
                        ),
                    ],
                ]
            ),
            wr.HorizontalRule(),
            wr.H1(text=["Getting Started"]),
            wr.P(text=["W&B has two components:"]),
            wr.OrderedList(
                items=[
                    ["A centrally managed MLOps platform and UI"],
                    [
                        "The ",
                        wr.InlineCode(code="wandb"),
                        " SDK (",
                        wr.Link(text="github", url="https://github.com/wandb/client"),
                        ", ",
                        wr.Link(text="pypi", url="https://pypi.org/project/wandb/"),
                        ", ",
                        wr.Link(
                            text="conda-forge",
                            url="https://anaconda.org/conda-forge/wandb",
                        ),
                        ")",
                    ],
                ]
            ),
            wr.P(),
            wr.H3(text=["1. Install the SDK"]),
            wr.CodeBlock(code=["pip install wandb"], language="python"),
            wr.P(),
            wr.H3(text=["2. Log in to W&B"]),
            wr.P(text=["You will be prompted to get and set your API key in the UI."]),
            wr.CodeBlock(code=["wandb.login()"], language="python"),
            wr.P(),
            wr.H3(text=["3. Setup an experiment"]),
            wr.P(
                text=[
                    "Add this to the beginning of your scripts (or top of your notebook)."
                ]
            ),
            wr.CodeBlock(code=["wandb.init()"], language="python"),
            wr.P(),
            wr.P(
                text=[
                    wr.Link(
                        text="For more details on options and advanced usage, see the docs.",
                        url="https://docs.wandb.ai/ref/python/init",
                    )
                ]
            ),
            wr.P(),
            wr.H3(text=["4. Log anything!"]),
            wr.P(text=["You can log metrics anywhere in your script, for example"]),
            wr.CodeBlock(code=['wandb.log({"loss": model_loss})'], language="python"),
            wr.P(),
            wr.P(
                text=[
                    "Log metrics, graphs, dataframes, images with segmentation masks or bounding boxes, videos, point clouds, custom HTML, and more!  ",
                    wr.Link(
                        text="For more details on logging, including advanced types, see the docs.",
                        url="https://docs.wandb.ai/guides/track/log",
                    ),
                ]
            ),
            wr.P(text=["W&B also helps you reproduce results by capturing:"]),
            wr.UnorderedList(
                items=[
                    ["git state (repo, commit)"],
                    ["requirements (requirements.txt, conda_env.yml)"],
                    ["logs, including stdout"],
                    [
                        "hardware metrics (CPU, GPU, network, memory utilization, temperature, throughput)"
                    ],
                    ["and more!"],
                ]
            ),
            wr.P(),
            wr.H3(text=["Putting everything together:"]),
            wr.CodeBlock(
                code=[
                    "wandb.login()",
                    "",
                    "wandb.init()",
                    "for i in range(1000):",
                    '    wandb.log({"metric": i})',
                ],
                language="python",
            ),
            wr.P(),
            wr.H1(text=["What else is possible with W&B?"]),
            wr.H2(text=["Example projects"]),
            wr.Gallery(
                ids=[
                    "Vmlldzo4ODc0MDc=",
                    "Vmlldzo4NDI3NzM=",
                    "Vmlldzo2MDIzMTg=",
                    "VmlldzoyMjA3MjY=",
                    "Vmlldzo1NjM4OA==",
                ]
            ),
            wr.P(),
        ],
    )



"""Batching file prepare requests to our API."""

import concurrent.futures
import functools
import os
import queue
import shutil
import threading
from typing import TYPE_CHECKING, NamedTuple, Optional, Union, cast

from wandb.filesync import dir_watcher, step_upload
from wandb.sdk.lib import filesystem, runid

if TYPE_CHECKING:
    import tempfile

    from wandb.filesync import stats
    from wandb.sdk.interface import artifacts
    from wandb.sdk.internal import artifact_saver, internal_api


class RequestUpload(NamedTuple):
    path: str
    save_name: dir_watcher.SaveName
    copy: bool


class RequestStoreManifestFiles(NamedTuple):
    manifest: "artifacts.ArtifactManifest"
    artifact_id: str
    save_fn: "artifact_saver.SaveFn"
    save_fn_async: "artifact_saver.SaveFnAsync"


class RequestCommitArtifact(NamedTuple):
    artifact_id: str
    finalize: bool
    before_commit: step_upload.PreCommitFn
    result_future: "concurrent.futures.Future[None]"


class RequestFinish(NamedTuple):
    callback: Optional[step_upload.OnRequestFinishFn]


Event = Union[
    RequestUpload, RequestStoreManifestFiles, RequestCommitArtifact, RequestFinish
]


class StepChecksum:
    def __init__(
        self,
        api: "internal_api.Api",
        tempdir: "tempfile.TemporaryDirectory",
        request_queue: "queue.Queue[Event]",
        output_queue: "queue.Queue[step_upload.Event]",
        stats: "stats.Stats",
    ) -> None:
        self._api = api
        self._tempdir = tempdir
        self._request_queue = request_queue
        self._output_queue = output_queue
        self._stats = stats

        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True

    def _thread_body(self) -> None:
        while True:
            req = self._request_queue.get()
            if isinstance(req, RequestUpload):
                path = req.path
                if req.copy:
                    path = os.path.join(
                        self._tempdir.name,
                        f"{runid.generate_id()}-{req.save_name}",
                    )
                    filesystem.mkdir_exists_ok(os.path.dirname(path))
                    try:
                        # certain linux distros throw an exception when copying
                        # large files: https://bugs.python.org/issue43743
                        shutil.copy2(req.path, path)
                    except OSError:
                        shutil._USE_CP_SENDFILE = False  # type: ignore[attr-defined]
                        shutil.copy2(req.path, path)
                self._stats.init_file(req.save_name, os.path.getsize(path))
                self._output_queue.put(
                    step_upload.RequestUpload(
                        path,
                        req.save_name,
                        None,
                        None,
                        req.copy,
                        None,
                        None,
                        None,
                    )
                )
            elif isinstance(req, RequestStoreManifestFiles):
                for entry in req.manifest.entries.values():
                    if entry.local_path:
                        self._stats.init_file(
                            entry.local_path,
                            cast(int, entry.size),
                            is_artifact_file=True,
                        )
                        self._output_queue.put(
                            step_upload.RequestUpload(
                                entry.local_path,
                                dir_watcher.SaveName(
                                    entry.path
                                ),  # typecast might not be legit
                                req.artifact_id,
                                entry.digest,
                                False,
                                functools.partial(req.save_fn, entry),
                                functools.partial(req.save_fn_async, entry),
                                entry.digest,
                            )
                        )
            elif isinstance(req, RequestCommitArtifact):
                self._output_queue.put(
                    step_upload.RequestCommitArtifact(
                        req.artifact_id,
                        req.finalize,
                        req.before_commit,
                        req.result_future,
                    )
                )
            elif isinstance(req, RequestFinish):
                break
            else:
                raise Exception("internal error")

        self._output_queue.put(step_upload.RequestFinish(req.callback))

    def start(self) -> None:
        self._thread.start()

    def is_alive(self) -> bool:
        return self._thread.is_alive()

    def finish(self) -> None:
        self._request_queue.put(RequestFinish(None))



import inspect
import re
import urllib
from typing import List as LList
from typing import Optional, Union

from ... import __version__ as wandb_ver
from ... import termwarn
from ..public import Api as PublicApi
from ._panels import UnknownPanel, WeavePanel, panel_mapping, weave_panels
from .runset import Runset
from .util import (
    Attr,
    Base,
    Block,
    InlineCode,
    InlineLaTeX,
    Link,
    Panel,
    coalesce,
    fix_collisions,
    nested_get,
    nested_set,
    weave_inputs,
)
from .validators import OneOf, TypeValidator


class UnknownBlock(Block):
    pass


class PanelGrid(Block):
    runsets: list = Attr(
        json_path="spec.metadata.runSets",
        validators=[TypeValidator(Runset, how="keys")],
    )
    panels: list = Attr(
        json_path="spec.metadata.panelBankSectionConfig.panels",
        validators=[TypeValidator(Panel, how="keys")],
    )
    custom_run_colors: dict = Attr(
        json_path="spec.metadata.customRunColors",
        validators=[
            TypeValidator(Union[str, tuple], how="keys"),
            TypeValidator(str, how="values"),
        ],
    )
    active_runset: Union[str, None] = Attr(json_path="spec.metadata.openRunSet")

    def __init__(
        self,
        runsets=None,
        panels=None,
        custom_run_colors=None,
        active_runset=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._spec = self._default_panel_grid_spec()
        self.runsets = coalesce(runsets, self._default_runsets())
        self.panels = coalesce(panels, self._default_panels())
        self.custom_run_colors = coalesce(custom_run_colors, {})
        self.active_runset = active_runset

    @active_runset.getter
    def active_runset(self):
        json_path = self._get_path("active_runset")
        index = nested_get(self, json_path)
        if index is None:
            return None
        else:
            return self.runsets[index].name

    @active_runset.setter
    def active_runset(self, name):
        json_path = self._get_path("active_runset")
        index = None
        for i, rs in enumerate(self.runsets):
            if rs.name == name:
                index = i
                break
        nested_set(self, json_path, index)

    @panels.getter
    def panels(self):
        json_path = self._get_path("panels")
        specs = nested_get(self, json_path)
        panels = []
        for pspec in specs:
            cls = panel_mapping.get(pspec["viewType"], UnknownPanel)
            if cls is UnknownPanel:
                termwarn(
                    inspect.cleandoc(
                        f"""
                        UNKNOWN PANEL DETECTED
                            This can happen if we have added new panels, but you are using an older version of the SDK.
                            If your report is loading normally, you can safely ignore this message (but we recommend not touching UnknownPanel)
                            If you think this is an error, please file a bug report including your SDK version ({wandb_ver}) and this spec ({pspec})
                        """
                    )
                )
            if cls is WeavePanel:
                for cls in weave_panels:
                    try:
                        cls.from_json(pspec)
                    except Exception:
                        pass
                    else:
                        break
            panels.append(cls.from_json(pspec))
        return panels

    @panels.setter
    def panels(self, new_panels):
        json_path = self._get_path("panels")
        new_specs = [p.spec for p in fix_collisions(new_panels)]
        nested_set(self, json_path, new_specs)

    @runsets.getter
    def runsets(self):
        json_path = self._get_path("runsets")
        specs = nested_get(self, json_path)
        return [Runset.from_json(spec) for spec in specs]

    @runsets.setter
    def runsets(self, new_runsets):
        json_path = self._get_path("runsets")
        new_specs = [rs.spec for rs in new_runsets]
        nested_set(self, json_path, new_specs)

    @custom_run_colors.getter
    def custom_run_colors(self):
        json_path = self._get_path("custom_run_colors")
        id_colors = nested_get(self, json_path)

        def is_groupid(s):
            for rs in self.runsets:
                if rs.spec["id"] in s:
                    return True
            return False

        def groupid_to_ordertuple(groupid):
            rs = self.runsets[0]
            if "-run:" in groupid:
                id, rest = groupid.split("-run:", 1)
            else:
                id, rest = groupid.split("-", 1)
            kvs = rest.split("-")
            kvs = [rs.pm_query_generator.pc_back_to_front(v) for v in kvs]
            keys, ordertuple = zip(*[kv.split(":") for kv in kvs])
            rs_name = self._get_rs_by_id(id).name
            return (rs_name, *ordertuple)

        def run_id_to_name(id):
            for rs in self.runsets:
                try:
                    run = PublicApi().run(f"{rs.entity}/{rs.project}/{id}")
                except Exception:
                    pass
                else:
                    return run.name
            raise ValueError("Unable to find this run!")

        color_settings = {}
        for id, c in id_colors.items():
            if id == "ref":
                continue
            if is_groupid(id):
                key = groupid_to_ordertuple(id)
            else:
                key = run_id_to_name(id)
            color_settings[key] = c
        return color_settings

    @custom_run_colors.setter
    def custom_run_colors(self, new_custom_run_colors):
        json_path = self._get_path("custom_run_colors")
        color_settings = {}

        def ordertuple_to_groupid(ordertuple):
            rs_name, rest = ordertuple[0], ordertuple[1:]
            rs = self._get_rs_by_name(rs_name)
            id = rs.spec["id"]
            keys = [rs.pm_query_generator.pc_front_to_back(k) for k in rs.groupby]
            kvs = [f"{k}:{v}" for k, v in zip(keys, rest)]
            linked = "-".join(kvs)
            return f"{id}-{linked}"

        def run_name_to_id(name):
            for rs in self.runsets:
                runs = PublicApi().runs(
                    path=f"{rs.entity}/{rs.project}", filters={"display_name": name}
                )
                if len(runs) > 1:
                    termwarn(
                        "Multiple runs with the same name found! Using the first one."
                    )
                for run in runs:
                    if run.name == name:
                        return run.id
            raise ValueError("Unable to find this run!")

        for name, c in new_custom_run_colors.items():
            if isinstance(name, tuple):
                key = ordertuple_to_groupid(name)
            else:
                key = run_name_to_id(name)
            color_settings[key] = c
        nested_set(self, json_path, color_settings)

    def _get_rs_by_id(self, id):
        for rs in self.runsets:
            if rs.spec["id"] == id:
                return rs

    def _get_rs_by_name(self, name):
        for rs in self.runsets:
            if rs.name == name:
                return rs

    @staticmethod
    def _default_panel_grid_spec():
        return {
            "type": "panel-grid",
            "children": [{"text": ""}],
            "metadata": {
                "openViz": True,
                "panels": {
                    "views": {"0": {"name": "Panels", "defaults": [], "config": []}},
                    "tabs": ["0"],
                },
                "panelBankConfig": {
                    "state": 0,
                    "settings": {
                        "autoOrganizePrefix": 2,
                        "showEmptySections": False,
                        "sortAlphabetically": False,
                    },
                    "sections": [
                        {
                            "name": "Hidden Panels",
                            "isOpen": False,
                            "panels": [],
                            "type": "flow",
                            "flowConfig": {
                                "snapToColumns": True,
                                "columnsPerPage": 3,
                                "rowsPerPage": 2,
                                "gutterWidth": 16,
                                "boxWidth": 460,
                                "boxHeight": 300,
                            },
                            "sorted": 0,
                            "localPanelSettings": {
                                "xAxis": "_step",
                                "smoothingWeight": 0,
                                "smoothingType": "exponential",
                                "ignoreOutliers": False,
                                "xAxisActive": False,
                                "smoothingActive": False,
                            },
                        }
                    ],
                },
                "panelBankSectionConfig": {
                    "name": "Report Panels",
                    "isOpen": False,
                    "panels": [],
                    "type": "grid",
                    "flowConfig": {
                        "snapToColumns": True,
                        "columnsPerPage": 3,
                        "rowsPerPage": 2,
                        "gutterWidth": 16,
                        "boxWidth": 460,
                        "boxHeight": 300,
                    },
                    "sorted": 0,
                    "localPanelSettings": {
                        "xAxis": "_step",
                        "smoothingWeight": 0,
                        "smoothingType": "exponential",
                        "ignoreOutliers": False,
                        "xAxisActive": False,
                        "smoothingActive": False,
                    },
                },
                "customRunColors": {},
                "runSets": [],
                "openRunSet": 0,
                "name": "unused-name",
            },
        }

    @staticmethod
    def _default_runsets():
        return [Runset()]

    @staticmethod
    def _default_panels():
        return []


class List(Base):
    @classmethod
    def from_json(cls, spec: dict) -> "Union[CheckedList, OrderedList, UnorderedList]":
        items = []
        for item in spec["children"]:
            text = []
            for elem in item["children"][0]["children"]:
                if elem.get("type") == "latex":
                    text.append(InlineLaTeX(elem["content"]))
                elif elem.get("type") == "link":
                    text.append(Link(elem["children"][0]["text"], elem["url"]))
                elif elem.get("inlineCode"):
                    text.append(InlineCode(elem["text"]))
                elif elem.get("text"):
                    text.append(elem["text"])
            items.append(text)
        checked = [item.get("checked") for item in spec["children"]]
        ordered = spec.get("ordered")

        # NAND: Either checked or ordered or neither (unordered), never both
        if all(x is None for x in checked):
            checked = None
        if checked is not None and ordered is not None:
            raise ValueError(
                "Lists can be checked, ordered or neither (unordered), but not both!"
            )

        if checked:
            return CheckedList(items, checked)
        elif ordered:
            return OrderedList(items)
        else:
            return UnorderedList(items)


class CheckedList(Block, List):
    items: list = Attr()
    checked: list = Attr()

    def __init__(self, items=None, checked=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if items is not None and checked is not None and len(items) != len(checked):
            raise ValueError("Items and checked lists must be the same length!")

        self.items = coalesce(items, [""])
        self.checked = coalesce(checked, [False for _ in self.items])

    @property
    def spec(self) -> dict:
        children = []
        for item, check in zip(self.items, self.checked):
            if isinstance(item, list):
                content = [
                    t.spec if not isinstance(t, str) else {"text": t} for t in item
                ]
            else:
                content = [{"text": item}]
            children.append(
                {
                    "type": "list-item",
                    "children": [{"type": "paragraph", "children": content}],
                    "checked": check,
                }
            )

        return {"type": "list", "children": children}


class OrderedList(Block, List):
    items: list = Attr()

    def __init__(self, items=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = coalesce(items, [""])

    @property
    def spec(self) -> dict:
        children = []
        for item in self.items:
            if isinstance(item, list):
                content = [
                    t.spec if not isinstance(t, str) else {"text": t} for t in item
                ]
            else:
                content = [{"text": item}]
            children.append(
                {
                    "type": "list-item",
                    "children": [{"type": "paragraph", "children": content}],
                    "ordered": True,
                }
            )

        return {"type": "list", "ordered": True, "children": children}


class UnorderedList(Block, List):
    items: list = Attr()

    def __init__(self, items=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = coalesce(items, [""])

    @property
    def spec(self) -> dict:
        children = []
        for item in self.items:
            if isinstance(item, list):
                content = [
                    t.spec if not isinstance(t, str) else {"text": t} for t in item
                ]
            else:
                content = [{"text": item}]
            children.append(
                {
                    "type": "list-item",
                    "children": [{"type": "paragraph", "children": content}],
                }
            )

        return {"type": "list", "children": children}


class Heading(Base):
    @classmethod
    def from_json(cls, spec: dict) -> "Union[H1,H2,H3]":
        level = spec["level"]
        level_mapping = {1: H1, 2: H2, 3: H3}
        if level not in level_mapping:
            raise ValueError(f"`level` must be one of {list(level_mapping.keys())}")

        if isinstance(spec["children"], str):
            text = spec["children"]
        else:
            text = []
            for elem in spec["children"]:
                if elem.get("type") == "latex":
                    text.append(InlineLaTeX(elem["content"]))
                elif elem.get("type") == "link":
                    text.append(Link(elem["children"][0]["text"], elem["url"]))
                elif elem.get("inlineCode"):
                    text.append(InlineCode(elem["text"]))
                elif elem.get("text"):
                    text.append(elem["text"])
        if not isinstance(text, list):
            text = [text]
        return level_mapping[level](text)


class H1(Block, Heading):
    text: Union[str, list, Link] = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    @property
    def spec(self) -> dict:
        if isinstance(self.text, list):
            content = [
                t.spec if not isinstance(t, str) else {"text": t} for t in self.text
            ]
        else:
            content = [{"text": self.text}]
        return {
            "type": "heading",
            "children": content,
            "level": 1,
        }


class H2(Block, Heading):
    text: Union[str, list, Link] = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    @property
    def spec(self) -> dict:
        if isinstance(self.text, list):
            content = [
                t.spec if not isinstance(t, str) else {"text": t} for t in self.text
            ]
        else:
            content = [{"text": self.text}]
        return {
            "type": "heading",
            "children": content,
            "level": 2,
        }


class H3(Block, Heading):
    text: Union[str, list, Link] = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    @property
    def spec(self) -> dict:
        if isinstance(self.text, list):
            content = [
                t.spec if not isinstance(t, str) else {"text": t} for t in self.text
            ]
        else:
            content = [{"text": self.text}]
        return {
            "type": "heading",
            "children": content,
            "level": 3,
        }


class BlockQuote(Block):
    text: str = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    @classmethod
    def from_json(cls, spec: dict) -> "BlockQuote":
        text = spec["children"][0]["text"]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {"type": "block-quote", "children": [{"text": self.text}]}


class CalloutBlock(Block):
    text: Union[str, list] = Attr()

    def __init__(self, text=" ", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    def __post_init__(self) -> None:
        if isinstance(self.text, str):
            self.text = self.text.split("\n")

    @classmethod
    def from_json(cls, spec: dict) -> "CalloutBlock":
        text = [child["children"][0]["text"] for child in spec["children"]]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {
            "type": "callout-block",
            "children": [
                {"type": "callout-line", "children": [{"text": text}]}
                for text in self.text
            ],
        }


class CodeBlock(Block):
    code: Union[str, list] = Attr()
    language: str = Attr()

    def __init__(self, code=" ", language=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code = code
        self.language = coalesce(language, "python")

    def __post_init__(self) -> None:
        if isinstance(self.code, str):
            self.code = self.code.split("\n")

    @classmethod
    def from_json(cls, spec: dict) -> "CodeBlock":
        code = [child["children"][0]["text"] for child in spec["children"]]
        language = spec.get("language", "python")
        return cls(code, language)

    @property
    def spec(self) -> dict:
        language = self.language.lower()
        return {
            "type": "code-block",
            "children": [
                {
                    "type": "code-line",
                    "children": [{"text": text}],
                    "language": language,
                }
                for text in self.code
            ],
            "language": language,
        }


class MarkdownBlock(Block):
    text: Union[str, list] = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text
        if isinstance(self.text, list):
            self.text = "\n".join(self.text)

    @classmethod
    def from_json(cls, spec: dict) -> "MarkdownBlock":
        text = spec["content"]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {
            "type": "markdown-block",
            "children": [{"text": ""}],
            "content": self.text,
        }


class LaTeXBlock(Block):
    text: Union[str, list] = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text
        if isinstance(self.text, list):
            self.text = "\n".join(self.text)

    @classmethod
    def from_json(cls, spec: dict) -> "LaTeXBlock":
        text = spec["content"]
        return cls(text)

    @property
    def spec(self) -> dict:
        return {
            "type": "latex",
            "children": [{"text": ""}],
            "content": self.text,
            "block": True,
        }


class Gallery(Block):
    ids: list = Attr(validators=[TypeValidator(str, how="keys")])

    def __init__(self, ids, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ids = ids

    @classmethod
    def from_json(cls, spec: dict) -> "Gallery":
        ids = spec["ids"]
        return cls(ids)

    @classmethod
    def from_report_urls(cls, urls: LList[str]) -> "Gallery":
        from .report import Report

        ids = [Report._url_to_report_id(url) for url in urls]
        return cls(ids)

    @property
    def spec(self) -> dict:
        return {"type": "gallery", "children": [{"text": ""}], "ids": self.ids}


class Image(Block):
    url: str = Attr()
    caption: Optional[str] = Attr()

    def __init__(self, url, caption=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
        self.caption = caption

    @classmethod
    def from_json(cls, spec: dict) -> "Image":
        url = spec["url"]
        caption = spec["children"][0]["text"] if spec.get("hasCaption") else None
        return cls(url, caption)

    @property
    def spec(self) -> dict:
        if self.caption:
            return {
                "type": "image",
                "children": [{"text": self.caption}],
                "url": self.url,
                "hasCaption": True,
            }
        else:
            return {"type": "image", "children": [{"text": ""}], "url": self.url}


class WeaveBlockSummaryTable(Block):
    """This is a hacky solution to support the most common way of getting Weave tables for now..."""

    entity: str = Attr()
    project: str = Attr()
    table_name: str = Attr()

    def __init__(self, entity, project, table_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.entity = entity
        self.project = project
        self.table_name = table_name

    @classmethod
    def from_json(cls, spec: dict) -> "WeaveBlockSummaryTable":
        entity = spec["config"]["panelConfig"]["exp"]["fromOp"]["inputs"]["obj"][
            "fromOp"
        ]["inputs"]["run"]["fromOp"]["inputs"]["project"]["fromOp"]["inputs"][
            "entityName"
        ][
            "val"
        ]
        project = spec["config"]["panelConfig"]["exp"]["fromOp"]["inputs"]["obj"][
            "fromOp"
        ]["inputs"]["run"]["fromOp"]["inputs"]["project"]["fromOp"]["inputs"][
            "projectName"
        ][
            "val"
        ]
        table_name = spec["config"]["panelConfig"]["exp"]["fromOp"]["inputs"]["key"][
            "val"
        ]
        return cls(entity, project, table_name)

    @property
    def spec(self) -> dict:
        return {
            "type": "weave-panel",
            "children": [{"text": ""}],
            "config": {
                "panelConfig": {
                    "exp": {
                        "nodeType": "output",
                        "type": {
                            "type": "tagged",
                            "tag": {
                                "type": "tagged",
                                "tag": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "entityName": "string",
                                        "projectName": "string",
                                    },
                                },
                                "value": {
                                    "type": "typedDict",
                                    "propertyTypes": {"project": "project"},
                                },
                            },
                            "value": {
                                "type": "list",
                                "objectType": {
                                    "type": "tagged",
                                    "tag": {
                                        "type": "typedDict",
                                        "propertyTypes": {"run": "run"},
                                    },
                                    "value": {
                                        "type": "union",
                                        "members": [
                                            {
                                                "type": "file",
                                                "extension": "json",
                                                "wbObjectType": {
                                                    "type": "table",
                                                    "columnTypes": {},
                                                },
                                            },
                                            "none",
                                        ],
                                    },
                                },
                            },
                        },
                        "fromOp": {
                            "name": "pick",
                            "inputs": {
                                "obj": {
                                    "nodeType": "output",
                                    "type": {
                                        "type": "tagged",
                                        "tag": {
                                            "type": "tagged",
                                            "tag": {
                                                "type": "typedDict",
                                                "propertyTypes": {
                                                    "entityName": "string",
                                                    "projectName": "string",
                                                },
                                            },
                                            "value": {
                                                "type": "typedDict",
                                                "propertyTypes": {"project": "project"},
                                            },
                                        },
                                        "value": {
                                            "type": "list",
                                            "objectType": {
                                                "type": "tagged",
                                                "tag": {
                                                    "type": "typedDict",
                                                    "propertyTypes": {"run": "run"},
                                                },
                                                "value": {
                                                    "type": "union",
                                                    "members": [
                                                        {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "_wandb": {
                                                                    "type": "typedDict",
                                                                    "propertyTypes": {
                                                                        "runtime": "number"
                                                                    },
                                                                }
                                                            },
                                                        },
                                                        {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "_step": "number",
                                                                "table": {
                                                                    "type": "file",
                                                                    "extension": "json",
                                                                    "wbObjectType": {
                                                                        "type": "table",
                                                                        "columnTypes": {},
                                                                    },
                                                                },
                                                                "_wandb": {
                                                                    "type": "typedDict",
                                                                    "propertyTypes": {
                                                                        "runtime": "number"
                                                                    },
                                                                },
                                                                "_runtime": "number",
                                                                "_timestamp": "number",
                                                            },
                                                        },
                                                        {
                                                            "type": "typedDict",
                                                            "propertyTypes": {},
                                                        },
                                                    ],
                                                },
                                            },
                                        },
                                    },
                                    "fromOp": {
                                        "name": "run-summary",
                                        "inputs": {
                                            "run": {
                                                "nodeType": "output",
                                                "type": {
                                                    "type": "tagged",
                                                    "tag": {
                                                        "type": "tagged",
                                                        "tag": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "entityName": "string",
                                                                "projectName": "string",
                                                            },
                                                        },
                                                        "value": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "project": "project"
                                                            },
                                                        },
                                                    },
                                                    "value": {
                                                        "type": "list",
                                                        "objectType": "run",
                                                    },
                                                },
                                                "fromOp": {
                                                    "name": "project-runs",
                                                    "inputs": {
                                                        "project": {
                                                            "nodeType": "output",
                                                            "type": {
                                                                "type": "tagged",
                                                                "tag": {
                                                                    "type": "typedDict",
                                                                    "propertyTypes": {
                                                                        "entityName": "string",
                                                                        "projectName": "string",
                                                                    },
                                                                },
                                                                "value": "project",
                                                            },
                                                            "fromOp": {
                                                                "name": "root-project",
                                                                "inputs": {
                                                                    "entityName": {
                                                                        "nodeType": "const",
                                                                        "type": "string",
                                                                        "val": self.entity,
                                                                    },
                                                                    "projectName": {
                                                                        "nodeType": "const",
                                                                        "type": "string",
                                                                        "val": self.project,
                                                                    },
                                                                },
                                                            },
                                                        }
                                                    },
                                                },
                                            }
                                        },
                                    },
                                },
                                "key": {
                                    "nodeType": "const",
                                    "type": "string",
                                    "val": self.table_name,
                                },
                            },
                        },
                        "__userInput": True,
                    }
                }
            },
        }


class WeaveBlockArtifact(Block):
    """This is a hacky solution to support the most common way of getting Weave artifacts for now..."""

    entity: str = Attr()
    project: str = Attr()
    artifact: str = Attr()
    tab: str = Attr(
        validators=[OneOf(["overview", "metadata", "usage", "files", "lineage"])]
    )

    def __init__(self, entity, project, artifact, tab="overview", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.entity = entity
        self.project = project
        self.artifact = artifact
        self.tab = tab

    @classmethod
    def from_json(cls, spec: dict) -> "WeaveBlockSummaryTable":
        inputs = weave_inputs(spec)
        entity = inputs["project"]["fromOp"]["inputs"]["entityName"]["val"]
        project = inputs["project"]["fromOp"]["inputs"]["projectName"]["val"]
        artifact = inputs["artifactName"]["val"]
        tab = spec["config"]["panelConfig"]["panelConfig"]["tabConfigs"]["overview"][
            "selectedTab"
        ]
        return cls(entity, project, artifact, tab)

    @property
    def spec(self) -> dict:
        return {
            "type": "weave-panel",
            "children": [{"text": ""}],
            "config": {
                "panelConfig": {
                    "exp": {
                        "nodeType": "output",
                        "type": {
                            "type": "tagged",
                            "tag": {
                                "type": "tagged",
                                "tag": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "entityName": "string",
                                        "projectName": "string",
                                    },
                                },
                                "value": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "project": "project",
                                        "artifactName": "string",
                                    },
                                },
                            },
                            "value": "artifact",
                        },
                        "fromOp": {
                            "name": "project-artifact",
                            "inputs": {
                                "project": {
                                    "nodeType": "output",
                                    "type": {
                                        "type": "tagged",
                                        "tag": {
                                            "type": "typedDict",
                                            "propertyTypes": {
                                                "entityName": "string",
                                                "projectName": "string",
                                            },
                                        },
                                        "value": "project",
                                    },
                                    "fromOp": {
                                        "name": "root-project",
                                        "inputs": {
                                            "entityName": {
                                                "nodeType": "const",
                                                "type": "string",
                                                "val": self.entity,
                                            },
                                            "projectName": {
                                                "nodeType": "const",
                                                "type": "string",
                                                "val": self.project,
                                            },
                                        },
                                    },
                                },
                                "artifactName": {
                                    "nodeType": "const",
                                    "type": "string",
                                    "val": self.artifact,
                                },
                            },
                        },
                        "__userInput": True,
                    },
                    "panelInputType": {
                        "type": "tagged",
                        "tag": {
                            "type": "tagged",
                            "tag": {
                                "type": "typedDict",
                                "propertyTypes": {
                                    "entityName": "string",
                                    "projectName": "string",
                                },
                            },
                            "value": {
                                "type": "typedDict",
                                "propertyTypes": {
                                    "project": "project",
                                    "artifactName": "string",
                                },
                            },
                        },
                        "value": "artifact",
                    },
                    "panelConfig": {
                        "tabConfigs": {"overview": {"selectedTab": self.tab}}
                    },
                }
            },
        }


class WeaveBlockArtifactVersionedFile(Block):
    """This is a hacky solution to support the most common way of getting Weave artifact verions for now..."""

    entity: str = Attr()
    project: str = Attr()
    artifact: str = Attr()
    version: str = Attr()
    file: str = Attr()

    def __init__(self, entity, project, artifact, version, file, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.entity = entity
        self.project = project
        self.artifact = artifact
        self.version = version
        self.file = file

    @classmethod
    def from_json(cls, spec: dict) -> "WeaveBlockSummaryTable":
        inputs = weave_inputs(spec)
        entity = inputs["artifactVersion"]["fromOp"]["inputs"]["project"]["fromOp"][
            "inputs"
        ]["entityName"]["val"]
        project = inputs["artifactVersion"]["fromOp"]["inputs"]["project"]["fromOp"][
            "inputs"
        ]["projectName"]["val"]
        artifact = inputs["artifactVersion"]["fromOp"]["inputs"]["artifactName"]["val"]
        version = inputs["artifactVersion"]["fromOp"]["inputs"]["artifactVersionAlias"][
            "val"
        ]
        file = inputs["path"]["val"]
        return cls(entity, project, artifact, version, file)

    @property
    def spec(self) -> dict:
        return {
            "type": "weave-panel",
            "children": [{"text": ""}],
            "config": {
                "panelConfig": {
                    "exp": {
                        "nodeType": "output",
                        "type": {
                            "type": "tagged",
                            "tag": {
                                "type": "tagged",
                                "tag": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "entityName": "string",
                                        "projectName": "string",
                                    },
                                },
                                "value": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "project": "project",
                                        "artifactName": "string",
                                        "artifactVersionAlias": "string",
                                    },
                                },
                            },
                            "value": {
                                "type": "file",
                                "extension": "json",
                                "wbObjectType": {"type": "table", "columnTypes": {}},
                            },
                        },
                        "fromOp": {
                            "name": "artifactVersion-file",
                            "inputs": {
                                "artifactVersion": {
                                    "nodeType": "output",
                                    "type": {
                                        "type": "tagged",
                                        "tag": {
                                            "type": "tagged",
                                            "tag": {
                                                "type": "typedDict",
                                                "propertyTypes": {
                                                    "entityName": "string",
                                                    "projectName": "string",
                                                },
                                            },
                                            "value": {
                                                "type": "typedDict",
                                                "propertyTypes": {
                                                    "project": "project",
                                                    "artifactName": "string",
                                                    "artifactVersionAlias": "string",
                                                },
                                            },
                                        },
                                        "value": "artifactVersion",
                                    },
                                    "fromOp": {
                                        "name": "project-artifactVersion",
                                        "inputs": {
                                            "project": {
                                                "nodeType": "output",
                                                "type": {
                                                    "type": "tagged",
                                                    "tag": {
                                                        "type": "typedDict",
                                                        "propertyTypes": {
                                                            "entityName": "string",
                                                            "projectName": "string",
                                                        },
                                                    },
                                                    "value": "project",
                                                },
                                                "fromOp": {
                                                    "name": "root-project",
                                                    "inputs": {
                                                        "entityName": {
                                                            "nodeType": "const",
                                                            "type": "string",
                                                            "val": self.entity,
                                                        },
                                                        "projectName": {
                                                            "nodeType": "const",
                                                            "type": "string",
                                                            "val": self.project,
                                                        },
                                                    },
                                                },
                                            },
                                            "artifactName": {
                                                "nodeType": "const",
                                                "type": "string",
                                                "val": self.artifact,
                                            },
                                            "artifactVersionAlias": {
                                                "nodeType": "const",
                                                "type": "string",
                                                "val": self.version,
                                            },
                                        },
                                    },
                                },
                                "path": {
                                    "nodeType": "const",
                                    "type": "string",
                                    "val": self.file,
                                },
                            },
                        },
                        "__userInput": True,
                    }
                }
            },
        }


class HorizontalRule(Block):
    @classmethod
    def from_json(cls, spec: dict) -> "HorizontalRule":
        return cls()

    @property
    def spec(self):
        return {"type": "horizontal-rule", "children": [{"text": ""}]}


class TableOfContents(Block):
    @classmethod
    def from_json(cls, spec: dict) -> "TableOfContents":
        return cls()

    @property
    def spec(self) -> dict:
        return {"type": "table-of-contents", "children": [{"text": ""}]}


class SoundCloud(Block):
    url: str = Attr()

    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url

    @classmethod
    def from_json(cls, spec: dict) -> "SoundCloud":
        quoted_url = spec["html"].split("url=")[-1].split("&show_artwork")[0]
        url = urllib.parse.unquote(quoted_url)
        return cls(url)

    @property
    def spec(self) -> dict:
        quoted_url = urllib.parse.quote(self.url)
        return {
            "type": "soundcloud",
            "html": f'<iframe width="100%" height="400" scrolling="no" frameborder="no" src="https://w.soundcloud.com/player/?visual=true&url={quoted_url}&show_artwork=true"></iframe>',
            "children": [{"text": ""}],
        }


class Twitter(Block):
    embed_html: str = Attr()

    def __init__(self, embed_html, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.embed_html = embed_html
        if self.embed_html:
            pattern = r" <script[\s\S]+?/script>"
            self.embed_html = re.sub(pattern, "\n", self.embed_html)

    @classmethod
    def from_json(cls, spec: dict) -> "Twitter":
        embed_html = spec["html"]
        return cls(embed_html)

    @property
    def spec(self) -> dict:
        return {"type": "twitter", "html": self.embed_html, "children": [{"text": ""}]}


class Spotify(Block):
    spotify_id: str = Attr()

    def __init__(self, spotify_id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spotify_id = spotify_id

    @classmethod
    def from_json(cls, spec: dict) -> "Spotify":
        return cls(spec["spotifyID"])

    @classmethod
    def from_url(cls, url: str) -> "Spotify":
        spotify_id = url.split("/")[-1].split("?")[0]
        return cls(spotify_id)

    @property
    def spec(self) -> dict:
        return {
            "type": "spotify",
            "spotifyType": "track",
            "spotifyID": self.spotify_id,
            "children": [{"text": ""}],
        }


class Video(Block):
    url: str = Attr()

    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url

    @classmethod
    def from_json(cls, spec: dict) -> "Video":
        return cls(spec["url"])

    @property
    def spec(self) -> dict:
        return {
            "type": "video",
            "url": self.url,
            "children": [{"text": ""}],
        }


class P(Block):
    text: Union[str, InlineLaTeX, InlineCode, Link, list] = Attr()

    def __init__(self, text="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = text

    @classmethod
    def from_json(cls, spec):
        if isinstance(spec["children"], str):
            text = spec["children"]
        else:
            text = []
            for elem in spec["children"]:
                if elem.get("type") == "latex":
                    text.append(InlineLaTeX(elem["content"]))
                elif elem.get("type") == "link":
                    text.append(Link(elem["children"][0]["text"], elem["url"]))
                elif elem.get("inlineCode"):
                    text.append(InlineCode(elem["text"]))
                elif elem.get("text"):
                    text.append(elem["text"])

        if not isinstance(text, list):
            text = [text]
        return cls(text)

    @property
    def spec(self) -> dict:
        if isinstance(self.text, list):
            content = [
                t.spec if not isinstance(t, str) else {"text": t} for t in self.text
            ]
        else:
            content = [{"text": self.text}]

        return {"type": "paragraph", "children": content}


class WeaveBlock(Block):
    def __init__(self, spec, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._spec = spec

    @classmethod
    def from_json(cls, spec):
        obj = cls(spec=spec)
        obj._spec = spec
        return obj

    @property
    def spec(self):
        return self._spec


block_mapping = {
    "block-quote": BlockQuote,
    "callout-block": CalloutBlock,
    "code-block": CodeBlock,
    "gallery": Gallery,
    "heading": Heading,
    "horizontal-rule": HorizontalRule,
    "image": Image,
    "latex": LaTeXBlock,
    "list": List,
    "markdown-block": MarkdownBlock,
    "panel-grid": PanelGrid,
    "paragraph": P,
    "table-of-contents": TableOfContents,
    "weave-panel": WeaveBlock,
    "video": Video,
    "spotify": Spotify,
    "twitter": Twitter,
    "soundcloud": SoundCloud,
}

weave_blocks = [
    WeaveBlockSummaryTable,
    WeaveBlockArtifactVersionedFile,
    WeaveBlockArtifact,
    WeaveBlock,
]
